from .core import FormatManager

pixel_formats = FormatManager()

from .core.functions import imread, mimread, imwrite, mimwrite
from .core.functions import get_reader, get_writer
from .core.functions import frame, empty, zeros, ones
from . import formats



from .gray import (Gray,
                   Gray10LE, Gray10BE,
                   Gray16LE, Gray16BE,
                   Gray9LE, Gray9BE,
                   Gray12LE, Gray12BE,
                   Gray14LE, Gray14BE)
from .yuv420p import (YUV420P,
                      YUV420P10LE, YUV420P10BE,
                      YUV420P16LE, YUV420P16BE,
                      YUV420P9LE, YUV420P9BE,
                      YUV420P12LE, YUV420P12BE,
                      YUV420P14LE, YUV420P14BE)
from .yuv422p import (YUV422P,
                      YUV422P10LE, YUV422P10BE,
                      YUV422P16LE, YUV422P16BE,
                      YUV422P9LE, YUV422P9BE,
                      YUV422P12LE, YUV422P12BE,
                      YUV422P14LE, YUV422P14BE)
from .yuv444p import (YUV444P,
                      YUV444P10LE, YUV444P10BE,
                      YUV444P16LE, YUV444P16BE,
                      YUV444P9LE, YUV444P9BE,
                      YUV444P12LE, YUV444P12BE,
                      YUV444P14LE, YUV444P14BE)
from .yuyv422 import (YUYV422, UYVY422, YVYU422)



from typing import Union, Tuple
import numpy as np


class YUVFrame:
    """YUVFrame provides convenient data access to a single yuv/ycbcr frame."""

    def __init__(self, data, pixel_format: str):
        self._data = data
        self._pixel_format = pixel_format

    @property
    def pixel_format(self):
        return self._pixel_format

    @property
    def y(self):
        return self._data['y']

    @y.setter
    def y(self, value: np.ndarray):
        self._data['y'] = value

    @property
    def u(self):
        return self._data['u']

    @u.setter
    def u(self, value: np.ndarray):
        self._data['u'] = value

    @property
    def v(self):
        return self._data['v']

    @v.setter
    def v(self, value: np.ndarray):
        self._data['v'] = value

    @property
    def cb(self):
        return self._data['u']

    @cb.setter
    def cb(self, value: np.ndarray):
        self._data['u'] = value

    @property
    def cr(self):
        return self._data['v']

    @cr.setter
    def cr(self, value: np.ndarray):
        self._data['v'] = value

    def __getitem__(self, item: Union[str, int]):
        if isinstance(item, int):
            item = ('y', 'u', 'v')[item]
        elif isinstance(item, str):
            if item == 'cb':
                item = 'u'
            elif item == 'cr':
                item = 'v'
        return self._data[item]

    def __setitem__(self, key: Union[str, int], value: np.ndarray):
        if isinstance(key, int):
            key = ('y', 'u', 'v')[key]
        elif isinstance(key, str):
            if key == 'cb':
                key = 'u'
            elif key == 'cr':
                key = 'v'
        self._data[key] = value

    def set(self, yuv: Tuple[np.ndarray, np.ndarray, np.ndarray]):
        self._data['y'] = yuv[0]
        self._data['u'] = yuv[1]
        self._data['v'] = yuv[2]

    def split(self):
        return self._data['y'], self._data['u'], self._data['v']



from io import IOBase
from pathlib import Path
import psutil
import numpy as np
from . import YUVFrame
from . import Format


class Reader:

    def __init__(self, file, format: Format):
        if isinstance(file, IOBase):
            self._file = file
        else:
            self._file = open(Path(file), 'rb')
        self._format = format
        self._length = self._length_from_filesize()
        self._iter_idx = 0

    def __len__(self):
        return self._length

    def __iter__(self):
        for i in range(self._length):
            yield self.read(i, count=1)[0]

    def _length_from_filesize(self):
        return Path(self._file.name).stat().st_size // self._format.dtype.itemsize

    def _validate_memory(self, count):
        available = psutil.virtual_memory().available
        required = count * self._format.dtype.itemsize
        if required > available * 0.9:
            raise RuntimeError("The required memory ({}) to read '{}' frames "
                               "from file '{}' exceeds 90% of the available system "
                               "memory ({})".format(required,
                                                    count,
                                                    self._file.name,
                                                    available))

    def read(self, index, count=None):
        if count is None:
            count = self._length
        if index + count > self._length:
            raise ValueError("Cannot read number of frames '{}' at index '{}' "
                             "from file '{}' with length '{}'.".format(count,
                                                                       index,
                                                                       self._file.name,
                                                                       self._length))
        self._validate_memory(count)
        self._file.seek(index * self._format.dtype.itemsize)
        unstructured_data = np.fromfile(self._file,
                                        self._format.dtype,
                                        count=count)
        y, u, v = self._format.unpack(unstructured_data)
        structured_data = np.empty(count,
                                   dtype=np.dtype([('y', y.dtype, y.shape[1:]),
                                                   ('u', u.dtype, u.shape[1:]),
                                                   ('v', v.dtype, v.shape[1:])]))
        structured_data['y'] = y
        structured_data['u'] = u
        structured_data['v'] = v
        yuv_frames = [YUVFrame(data, self._format.identifier()) for data in structured_data]
        return yuv_frames
